# -*- coding: utf-8 -*-
from struct import pack
from struct import unpack

from .constants import NO_KEY_DETECTED
from .internal import XidConnection
from .keymaps import (rb_530_keymap, rb_730_keymap, rb_830_keymap,
                      rb_834_keymap, lumina_keymap)

import ftd2xx

class XidScanner(object):
    """
    Scan the computer for connected XID devices
    """
    def __init__(self):
        self.__xid_cons = []
        self.detect_xid_devices()

    def detect_xid_devices(self):
        """
        For all of the com ports connected to the computer, send an
        XID command '_c1'.  If the device response with '_xid', it is
        an xid device.
        """
        self.__xid_cons = []

        devs = ftd2xx.listDevices()

        if devs is None:
            return

        for d in devs:
            device_found = False
            for b in [115200, 19200, 9600, 57600, 38400]:
                con = XidConnection(d, b)

                if con.open():
                    con.flush()
                
                    try:
                        returnval = con.send_xid_command("_c1", 5).decode('ASCII')
                    except UnicodeDecodeError as e:
                        # Assume this isn't an XID device, since it returned something weird.
                        con.close()
                        break

                    if returnval.startswith('_xid'):
                        device_found = True
                        self.__xid_cons.append(con)

                        if(returnval != '_xid0'):
                            # set the device into XID mode
                            con.send_xid_command('c10')
                            con.flush()

                    con.close()
                    if device_found:
                        # Device found, we're done.
                        break
                else:
                    continue

    def device_at_index(self, index):
        """
        Returns the device at the specified index
        """
        if index >= len(self.__xid_cons):
            raise ValueError("Invalid device index")

        return self.__xid_cons[index]

    def device_count(self):
        """
        Number of XID devices connected to the computer
        """
        return len(self.__xid_cons)


class XidError(Exception):
    pass


class XidDevice(object):
    """
    Class for interfacing with a Cedrus XID device.

    At the beginning of an experiment, the developer should call:

        XidDevice.reset_base_timer()

    Whenever a stimulus is presented, the developer should call:

        XidDevice.reset_rt_timer()
    """

    def __init__(self, xid_connection):
        self.con = xid_connection
        self._impl = None
        self.product_id = -1
        self.model_id = -1
        self.major_fw_version = -1
        self.device_name = 'Uninitialized XID device'
        self.keymap = None
        self.response_queue = []
        
        self.init_device()

        self.con.set_using_stim_tracker_output(self.major_fw_version == 2 or self.product_id == b'S')
        self.con.set_resp_packet_size(self.major_fw_version == 2 and self.product_id == b'S')
        if self.major_fw_version == 1:
            self.con.send_xid_command('a10')
        self.con.clear_digital_output_lines(0xff)

    def __del__(self):
        self.con.close()
        del self.con

    def reset_rt_timer(self):
        """
        Resets the Reaction Time timer.
        """
        self.con.send_xid_command("e5")

    def reset_base_timer(self):
        """
        Resets the base timer
        """
        if self.major_fw_version < 2:
            self.con.send_xid_command("e1")

    def query_base_timer(self):
        """
        gets the value from the device's base timer
        """
        time = -1
        if self.major_fw_version < 2:
            (_, _, time) = unpack('<ccI', self.con.send_xid_command("e3", 6))
        
        return time

    def is_response_device(self):
        """
        "Response device" is used loosely here. Second generation StimTrackers are not response
        devices in a strict sense, they are capable of reporting keypresses via USB.
        The only devices you will never get a keypress from are first gen StimTrackers and c-pods.
        """
        return not (self.product_id == b'S' and self.major_fw_version != 2) and self.product_id != b'4'

    def init_device(self):
        """
        Initializes the device with the proper keymaps and name
        """
        self.product_id = self._send_command('_d2', 1)
        self.model_id = self._send_command('_d3', 1)
        self.major_fw_version = int(self._send_command('_d4', 1))

        if self.product_id == b'0':
            self.device_name = 'Cedrus Lumina 3G' if self.major_fw_version == 2 else 'Cedrus Lumina LP-400'
            self.keymap = lumina_keymap
        elif self.product_id == b'1':
            self.device_name = 'Cedrus SV-1 Voice Key'
        elif self.product_id == b'2':
            if self.model_id == b'1':
                self.device_name = 'Cedrus RB-540' if self.major_fw_version == 2 else 'Cedrus RB-530'
                self.keymap = rb_530_keymap
            elif self.model_id == b'2':
                self.device_name = 'Cedrus RB-740' if self.major_fw_version == 2 else 'Cedrus RB-730'
                self.keymap = rb_730_keymap
            elif self.model_id == b'3':
                self.device_name = 'Cedrus RB-840' if self.major_fw_version == 2 else 'Cedrus RB-830'
                self.keymap = rb_830_keymap
            elif self.model_id == b'4':
                self.device_name = 'Cedrus RB-844' if self.major_fw_version == 2 else 'Cedrus RB-834'
                self.keymap = rb_834_keymap
            else:
                raise XidError('Unknown RB Device')
        elif self.product_id == b'4':
            self.device_name = 'Cedrus C-POD'
        elif self.product_id == b'S':
            if self.major_fw_version < 2:
                self.device_name = 'Cedrus StimTracker'
            else:
                self.device_name = 'Cedrus StimTracker Duo' if self.model_id == b'1' else 'Cedrus StimTracker Quad'

        elif self.product_id == -99:
            raise XidError('Invalid XID device')

    def _send_command(self, command, expected_bytes):
        """
        Send an XID command to the device
        """
        response = self.con.send_xid_command(command, expected_bytes)

        return response

    def poll_for_response(self):
        """
        Polls the device for user input

        If there is a keymapping for the device, the key map is applied
        to the key reported from the device. This only applies to port 0
        (typically the physical buttons) responses. Rest are unchanged.

        If a response is waiting to be processed, the response is appended
        to the internal response_queue
        """
        key_state = self.con.check_for_keypress()

        if key_state != NO_KEY_DETECTED:
            response = self.con.get_current_response()

            if response['port'] == 0:
                if self.keymap is not None:
                    response['key'] = self.keymap[response['key']]
                else:
                    response['key'] -= 1

            self.response_queue.append(response)

    def response_queue_size(self):
        """
        Number of responses in the response queue
        """
        return len(self.response_queue)

    def has_response(self):
        """
        Do we have responses in the queue
        """
        return len(self.response_queue) > 0

    def get_next_response(self):
        """
        Pops the response at the beginning of the response queue
        and returns it.

        This function returns a dict object with the following keys:

            pressed:  A boolean value of whether the event was a keypress
                      or key release.
            key:      The key on the device that was pressed.  This is a
                      0 based index.
            port:     Device port the response came from.  Typically this
                      is 0 on RB-series devices, and 2 on SV-1 voice key
                      devices.
            time:     For the time being, this just returns 0.  There is
                      currently an issue with clock drift in the Cedrus XID
                      devices.  Once we have this issue resolved, time will
                      report the value of the RT timer in miliseconds.
        """
        response = None
        if self.has_response():
            response = self.response_queue.pop(0)
        return response

    def clear_response_queue(self):
        """
        Clears the response queue
        """
        self.response_queue = []

    # Will flush both input and output buffers by default.
    # 1 is output (from device) only, 2 is input (to device) only
    def flush_serial_buffer(self, mask=0):
        self.con.flush(mask)

    def set_pulse_duration(self, duration):
        """
        Sets the pulse duration for events in miliseconds when activate_line
        is called
        """
        if duration > 4294967295:
            raise ValueError('Duration is too long. Please choose a value '
                             'less than 4294967296.')

        big_endian = hex(duration)[2:]
        if len(big_endian) % 2 != 0:
            big_endian = '0'+big_endian

        little_endian = []

        for i in range(0, len(big_endian), 2):
            little_endian.insert(0, big_endian[i:i+2])

        for i in range(0, 4-len(little_endian)):
            little_endian.append('00')

        command = 'mp'
        for i in little_endian:
            command += chr(int(i, 16))

        self.con.send_xid_command(command, 0)

    def activate_line(self, lines=None, bitmask=None, leave_remaining_lines=False):
        """
        Triggers an output line.

        There are up to 16 output lines on XID devices that can be raised
        in any combination.  To raise lines 1 and 7, for example, you pass
        in the list: activate_line(lines=[1, 7]).

        To raise a single line, pass in just an integer, or a list with a
        single element to the lines keyword argument:

            activate_line(lines=3)

            or

            activate_line(lines=[3])

        The `lines` argument must either be an Integer, list of Integers, or
        None.

        If you'd rather specify a bitmask for setting the lines, you can use
        the bitmask keyword argument.  Bitmask must be a Integer value between
        0 and 255 where 0 specifies no lines, and 255 is all lines.  For a
        mapping between lines and their bit values, see the `_lines` class
        variable.

        To use this, call the function as so to activate lines 1 and 6:

            activate_line(bitmask=33)

        leave_remaining_lines tells the function to only operate on the lines
        specified.  For example, if lines 1 and 8 are active, and you make
        the following function call:

            activate_line(lines=4, leave_remaining_lines=True)

        This will result in lines 1, 4 and 8 being active.

        If you call activate_line(lines=4) with leave_remaining_lines=False
        (the default), if lines 1 and 8 were previously active, only line 4
        will be active after the call.
        """
        if lines is None and bitmask is None:
            raise ValueError('Must set one of lines or bitmask')
        if lines is not None and bitmask is not None:
            raise ValueError('Can only set one of lines or bitmask')

        if bitmask is not None:
            if bitmask not in list(range(0, 65536)):
                raise ValueError('bitmask must be an integer between '
                                 '0 and 65535')

        if lines is not None:
            if not isinstance(lines, list):
                lines = [lines]

            bitmask = 0
            for l in lines:
                if l < 1 or l > 16:
                    raise ValueError('Line numbers must be between 1 and 16 '
                                     '(inclusive)')
                bitmask |= 2 ** (l-1)

        self.con.set_digital_output_lines(bitmask, leave_remaining_lines)

    # A more straightforward function for raising lines. You tell it the mask you want to see
    # and it sets it.
    def set_lines(self, lines):
        self.con.set_digio_lines_to_mask(lines)

    def clear_line(self, lines=None, bitmask=None, leave_remaining_lines=False):
        """
        The inverse of activate_line.  If a line is active, it deactivates it.

        This has the same parameters as activate_line()
        """
        if lines is None and bitmask is None:
            raise ValueError('Must set one of lines or bitmask')
        if lines is not None and bitmask is not None:
            raise ValueError('Can only set one of lines or bitmask')

        if bitmask is not None:
            if bitmask not in list(range(0, 65536)):
                raise ValueError('bitmask must be an integer between '
                                 '0 and 65535')

        if lines is not None:
            if not isinstance(lines, list):
                lines = [lines]

            bitmask = 0
            for l in lines:
                if l < 1 or l > 16:
                    raise ValueError('Line numbers must be between 1 and 16 '
                                     '(inclusive)')
                bitmask |= 2 ** (l-1)

        self.con.clear_digital_output_lines(bitmask, leave_remaining_lines)

    def enable_usb_output(self, selector, enable):
        if self.major_fw_version < 2:
            return

        command = 'iu%s%s' % (selector, '1' if enable is True else '0')
        self.con.send_xid_command(command, 0)

    def get_pulse_table_bitmask(self):
        lines = 0
        if self.major_fw_version > 1:
            (_, _, _, lines) = unpack('<cccH', self.con.send_xid_command("_mk", 5))

        return lines

    def set_pulse_table_bitmask(self, mask):
        ptable_mask = pack('<ccH', b'm', b'k', mask)
        if self.major_fw_version > 1:
            self.con.send_xid_byte_command(ptable_mask)

    def clear_pulse_table(self):
        if self.major_fw_version > 1:
            self.con.send_xid_command("mc")

    def is_pulse_table_running(self):
        running = -1
        if self.major_fw_version > 1:
            (_, _, _, running) = unpack('<cccc', self.con.send_xid_command("_mr", 4))

        return running == b'1'

    def run_pulse_table(self):
        if self.major_fw_version > 1:
            self.con.send_xid_command("mr")

    def stop_pulse_table(self):
        if self.major_fw_version > 1:
            self.con.send_xid_command("ms")

    def add_pulse_table_entry(self, time, mask):
        ptable_mask = pack('<ccIH', b'm', b't', time, mask)
        if self.major_fw_version > 1:
            self.con.send_xid_byte_command(ptable_mask)

    def reset_output_lines(self):
        if self.major_fw_version > 1:
            self.con.send_xid_command("mz")

    def __getattr__(self, attrname):
        return getattr(self._impl, attrname)

    def __str__(self):
        return '<XidDevice "%s">' % self.device_name

    def __repr__(self):
        return self.__str__()



# -*- coding: utf-8 -*-

from .pyxid_impl import *  # noqa

def get_xid_devices():
    """
    Returns a list of all Xid devices connected to your computer.
    """
    devices = []
    scanner = XidScanner()
    for i in range(scanner.device_count()):
        com = scanner.device_at_index(i)
        if com.open():
            device = XidDevice(com)

            device.reset_base_timer()
            device.reset_rt_timer()

            devices.append(device)
        else:
            continue
    return devices


def get_xid_device(device_number):
    """
    returns device at a given index.

    Raises ValueError if the device at the passed in index doesn't
    exist.
    """
    scanner = XidScanner()
    com = scanner.device_at_index(device_number)
    com.open()
    return XidDevice(com)


def test_event_loop(devices):
    for d in devices:
        d.reset_base_timer()
        d.reset_rt_timer()

    while True:
        for d in devices:
            if d.is_response_device():
                d.poll_for_response()

                if d.response_queue_size() > 0:
                    print(d.device_name, d.get_next_response())



# -*- coding: utf-8 -*-
from struct import pack
from struct import unpack
import sys, time
from .constants import NO_KEY_DETECTED, FOUND_KEY_DOWN, FOUND_KEY_UP, \
     KEY_RELEASE_BITMASK, INVALID_PORT_BITS, XID_PACKET_SIZE, ST2_PACKET_SIZE

try:
    import ftd2xx
except OSError as e:
    if 'image not found' in str(e):
        raise OSError('ftd2xx drivers are not installed (or not in expected location)'
                      ' and these are required for the Cedrus pyxid2 library.\n'
                      '** Download from https://www.ftdichip.com/Drivers/D2XX.htm **')
    else:
        raise(e)  # not an error we know so pass it on



class XidConnection(object):
    def __init__(self, device, baud_rate=115200):
        self.ftd2xx_intermediate = device
        self.ftd2xx_con = 0
        self.baudrate = baud_rate
        self.__needs_interbyte_delay = True
        self.__packet_size = XID_PACKET_SIZE
        self.__response_buffer = b''
        self.__response_structs_queue = []
        # The set lines cmd on XID 1 response devices (RB-x30 series, Lumina LP-400 and SV-1)'ah'.
        # In all other cases (ST-1, XID2 devices) 'mh' is used instead.
        self.__using_stim_tracker = False
        self.__set_lines_cmd = 'ah'+chr(0)+chr(0)
        self.__line_state = 0

    def set_using_stim_tracker_output(self, using_st=True):
        if using_st:
            self.__using_stim_tracker = True
            self.__set_lines_cmd = 'mh'+chr(0)+chr(0)
            self.__needs_interbyte_delay = False
        else:
            self.__using_stim_tracker = False
            self.__set_lines_cmd = 'ah'+chr(0)+chr(0)
            self.__needs_interbyte_delay = True

    def set_resp_packet_size(self, st2_packet_size=True):
        if st2_packet_size:
            self.__packet_size = ST2_PACKET_SIZE # ST2 packets are larger

    def clear_digital_output_lines(self, lines, leave_remaining_lines=False):
        if lines not in list(range(0, 65536)):
            raise ValueError('lines must be between 0 and 65535')

        local_lines = ~lines
        if local_lines < 0:
            local_lines += 65536

        self.set_digital_output_lines(local_lines, leave_remaining_lines)

    def set_digital_output_lines(self, lines, leave_remaining_lines=False):
        if lines not in list(range(0, 65536)):
            raise ValueError('lines must be between 0 and 65535')

        if leave_remaining_lines:
            lines |= self.__line_state

        if self.__using_stim_tracker:
            self.__set_lines_cmd = 'mh'+chr(lines & 0x000000FF)+chr((lines >> 8) & 0x000000FF)
        else:
            lines_tmp = ~lines
            if lines_tmp < 0:
                lines_tmp += 65536
            self.__set_lines_cmd = 'ah'+chr(lines_tmp & 0x000000FF)+chr((lines >> 8) & 0x000000FF)

        self.write(self.__set_lines_cmd)
        self.__line_state = lines

    def set_digio_lines_to_mask(self, lines):
        command_char = b'm' if self.__using_stim_tracker else b'a'

        digio_cmd = pack('<ccH', command_char, b'h', lines)
        self.write_bytes(digio_cmd)

    def flush(self, mask=0):
        self.ftd2xx_con.purge(mask)

    def open(self):
        try:
            self.ftd2xx_con = ftd2xx.openEx(self.ftd2xx_intermediate)
        except ftd2xx.DeviceError:
            return False

        self.ftd2xx_con.setBaudRate(self.baudrate)
        self.ftd2xx_con.setDataCharacteristics(8, 0, 0)

        self.ftd2xx_con.setTimeouts(50, 50)
        self.ftd2xx_con.setUSBParameters(64,64)
        self.ftd2xx_con.setLatencyTimer(10)
        self.flush()

        return True

    def close(self):
        self.ftd2xx_con.close()

    def send_xid_command(self, command, bytes_expected=0):
        self.write(command)

        response = self.read(bytes_expected)

        return response

    def send_xid_byte_command(self, command, bytes_expected=0):
        self.write_bytes(command)

        response = self.read(bytes_expected)

        return response

    def read(self, bytes_to_read):
        return self.ftd2xx_con.read(bytes_to_read)

    def set_timeout(self, timeout):
        self.ftd2xx_con.setTimeouts(timeout, 50)

    def write(self, command):
        bytes_written = 0
        cmd_bytes = []

        self.ftd2xx_con.setTimeouts(50, 50)

        for i in command:
            if (sys.version_info >= (3, 0)):
                cmd_bytes += i.encode('latin1')
            else:
                cmd_bytes += i

        if self.__needs_interbyte_delay:
            for char in cmd_bytes:
                bytes_written += self.ftd2xx_con.write(bytes([char]))
                time.sleep(0.001)
        else:
            bytes_written = self.ftd2xx_con.write(bytes(cmd_bytes))

        return bytes_written

    def write_bytes(self, command):
        bytes_written = 0

        self.ftd2xx_con.setTimeouts(50, 50)

        if self.__needs_interbyte_delay:
            for char in command:
                bytes_written += self.ftd2xx_con.write(char)
                time.sleep(0.001)
        else:
            bytes_written = self.ftd2xx_con.write(command)

        return bytes_written

    def check_for_keypress(self):
        self.ftd2xx_con.setTimeouts(2, 50)
        response = self.read(self.__packet_size)

        response_found = NO_KEY_DETECTED
        if len(response) > 0:
            self.__response_buffer += response
            if self.__packet_size == 6:
                response_found = self.xid_input_found()
            else:
                response_found = self.st2_input_found()

        return response_found

    def xid_input_found(self):
        input_found = NO_KEY_DETECTED

        position_in_buf = 0

        while ((position_in_buf + self.__packet_size) <=
               len(self.__response_buffer)):

            exception_free = True

            try:
                (k, params, time) = unpack('<cBI',
                                           self.__response_buffer[
                                               position_in_buf:
                                               (position_in_buf +
                                                self.__packet_size)])
            except Exception as exc:
                exception_free = False
                print(('Failed to unpack serial bytes in xid_input_found. '
                      'Err: ' + str(exc)))

            if exception_free:
                """
                Refer to PROTOCOL AND TIMING COMMANDS section of
                https://cedrus.com/support/xid/commands.htm
                """
                if (k != b'k' or (params & INVALID_PORT_BITS) != 0):
                    self.__response_buffer = b''
                    self.flush()
                    print('Pyxid found unparseable bytes in the buffer. '
                          'Flushing buffer.')

                    break
                else:
                    response = {'port': 0,
                                'key': 0,
                                'pressed': False,
                                'time': 0}
                    response['port'] = params & 0x0F
                    response['key'] = ((params & 0xE0) >> 5)
                    response['pressed'] = (params & KEY_RELEASE_BITMASK) == \
                        KEY_RELEASE_BITMASK

                    if response['key'] == 0:
                        response['key'] = 8

                    response['time'] = time

                    if response['pressed']:
                        input_found = FOUND_KEY_DOWN
                    else:
                        input_found = FOUND_KEY_UP

                    self.__response_structs_queue += [response]

            position_in_buf += self.__packet_size

        self.__response_buffer = self.__response_buffer[position_in_buf:]

        return input_found

    def st2_input_found(self):
        input_found = NO_KEY_DETECTED

        position_in_buf = 0

        while ((position_in_buf + self.__packet_size) <=
               len(self.__response_buffer)):

            exception_free = True

            try:
                (o, port, key, pressed, time, null_byte) = unpack('<ccBcIB',
                                           self.__response_buffer[
                                               position_in_buf:
                                               (position_in_buf +
                                                self.__packet_size)])
            except Exception as exc:
                exception_free = False
                print(('Failed to unpack serial bytes in st2_input_found. '
                      'Err: ' + str(exc)))

            if exception_free:
                """
                Refer to PROTOCOL AND TIMING COMMANDS section of
                https://cedrus.com/support/xid/commands.htm
                """
                if (o != b'o' or null_byte != 0):
                    self.__response_buffer = b''
                    self.flush()
                    print('Pyxid found unparseable bytes in the buffer. '
                          'Flushing buffer.')

                    break
                else:
                    response = {'port': 0,
                                'key': 0,
                                'pressed': False,
                                'time': 0}
                    response['port'] = port
                    response['key'] = key
                    response['pressed'] = True if pressed == b'1' else False

                    if response['key'] == 0:
                        response['key'] = 8

                    response['time'] = time

                    if response['pressed']:
                        input_found = FOUND_KEY_DOWN
                    else:
                        input_found = FOUND_KEY_UP

                    self.__response_structs_queue += [response]

            position_in_buf += self.__packet_size

        self.__response_buffer = self.__response_buffer[position_in_buf:]

        return input_found

    def get_current_response(self):
        """
        reads the current response data from the object and returns
        it in a dict.

        Currently 'time' is reported as 0 until clock drift issues are
        resolved.
        """
        response = {'port': 0,
                    'pressed': False,
                    'key': 0,
                    'time': 0}
        if len(self.__response_structs_queue) > 0:
            # make a copy just in case any other internal members of
            # XidConnection were tracking the structure
            response = self.__response_structs_queue[0].copy()
            # we will now hand over 'response' to the calling code,
            # so remove it from the internal queue
            self.__response_structs_queue.pop(0)

        return response



from setuptools import setup, find_packages

# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.txt'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name = "pyxid2",
    version = "1.0.4",
    packages = find_packages(),
    install_requires = ["ftd2xx>=1.1.2"],
    author = "Eugene Matsak",
    author_email = "developers@cedrus.com",
    maintainer = "Cedrus Corporation",
    maintainer_email = "developers@cedrus.com",
    description = ("Python library for interfacing with Cedrus XID devices, e.g. StimTracker, RB-x40, c-pod, and Lumina."),
    long_description = long_description,
    long_description_content_type='text/markdown',
    license = "BSD",
    keywords = "cedrus xid XID stimulus response data collection",
    url = "http://www.github.com/cedrus-opensource/pyxid/",
    classifiers = [
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Topic :: System :: Hardware",
        "Programming Language :: Python",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        ]
    )
